"""Support for Vodafone Station."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import UTC, datetime
from http import HTTPStatus
from http.cookies import SimpleCookie
from typing import Any

from aiohttp import (
    ClientResponse,
    ClientResponseError,
    ClientSession,
    ClientTimeout,
)
from yarl import URL

from .const import (
    _LOGGER,
    DEFAULT_TIMEOUT,
    HEADERS,
)
from .exceptions import (
    GenericResponseError,
)


@dataclass
class VodafoneStationDevice:
    """Vodafone Station device class."""

    connected: bool
    connection_type: str
    ip_address: str
    name: str
    mac: str
    type: str
    wifi: str


class VodafoneStationCommonApi(ABC):
    """Common API calls for Vodafone Station routers."""

    def __init__(
        self,
        url: URL,
        username: str,
        password: str,
        session: ClientSession,
    ) -> None:
        """Initialize the scanner."""
        self.username = username
        self.password = password
        self.base_url = url
        self.headers = HEADERS
        self.session = session
        self.csrf_token: str = ""
        self.encryption_key: str = ""
        self._unique_id: str | None = None
        self._overview: dict[str, Any] = {}
        self._devices: dict[str, VodafoneStationDevice] = {}

    async def _set_cookie(self) -> None:
        """Enable required session cookie."""
        self.session.cookie_jar.update_cookies(
            SimpleCookie(f"domain={self.base_url.host}; name=login_uid; value=1;"),
        )

    async def _request_page_result(
        self,
        method: str,
        page: str,
        payload: dict[str, Any] | None = None,
        timeout: ClientTimeout = DEFAULT_TIMEOUT,
    ) -> ClientResponse:
        """Request data from a web page."""
        _LOGGER.debug("%s page %s from host %s", method, page, self.base_url.host)
        timestamp = int(datetime.now(tz=UTC).timestamp())
        url = self.base_url.joinpath(page).with_query(
            _=timestamp, csrf_token=self.csrf_token
        )
        try:
            response = await self.session.request(
                method,
                url,
                data=payload,
                headers=self.headers,
                timeout=timeout,
                ssl=False,
                allow_redirects=True,
            )
            if response.status != HTTPStatus.OK:
                _LOGGER.warning(
                    "%s page %s from host %s failed: %s",
                    method,
                    page,
                    self.base_url.host,
                    response.status,
                )
                raise GenericResponseError
        except ClientResponseError as err:
            _LOGGER.exception(
                "%s page %s from host %s failed", method, page, self.base_url.host
            )
            raise GenericResponseError from err
        else:
            return response

    @abstractmethod
    def convert_uptime(self, uptime: str) -> datetime:
        """Convert uptime to datetime."""

    @abstractmethod
    async def login(self, force_logout: bool = False) -> bool:
        """Router login."""

    @abstractmethod
    async def get_devices_data(self) -> dict[str, VodafoneStationDevice]:
        """Get router device data."""

    @abstractmethod
    async def get_sensor_data(self) -> dict[str, Any]:
        """Get router sensor data."""

    @abstractmethod
    async def get_docis_data(self) -> dict[str, Any]:
        """Get router docis data."""

    @abstractmethod
    async def get_voice_data(self) -> dict[str, Any]:
        """Get router voice data."""

    @abstractmethod
    async def restart_connection(self, connection_type: str) -> None:
        """Internet Connection restart."""

    @abstractmethod
    async def restart_router(self) -> None:
        """Router restart."""

    @abstractmethod
    async def logout(self) -> None:
        """Router logout."""



"""Vodafone Station models package."""

from collections.abc import Mapping
from enum import Enum
from http import HTTPStatus
from typing import Any

import orjson
from aiohttp import (
    ClientConnectorError,
    ClientConnectorSSLError,
    ClientSession,
)
from yarl import URL

from aiovodafone.api import VodafoneStationCommonApi
from aiovodafone.const import _LOGGER, DEVICES_SETTINGS, HEADERS
from aiovodafone.exceptions import ModelNotSupported

from .sercomm import VodafoneStationSercommApi
from .technicolor import VodafoneStationTechnicolorApi
from .ultrahub import VodafoneStationUltraHubApi


class DeviceType(str, Enum):
    """Supported device types."""

    SERCOMM = "Sercomm"
    TECHNICOLOR = "Technicolor"
    ULTRAHUB = "UltraHub"


class_registry: dict[DeviceType, type[VodafoneStationCommonApi]] = {
    DeviceType.SERCOMM: VodafoneStationSercommApi,
    DeviceType.TECHNICOLOR: VodafoneStationTechnicolorApi,
    DeviceType.ULTRAHUB: VodafoneStationUltraHubApi,
}


def init_device_class(
    url: URL, device_type: DeviceType, data: Mapping[str, Any], session: ClientSession
) -> VodafoneStationCommonApi:
    """Return the inited API class."""
    if device_type not in class_registry:
        raise ModelNotSupported(f"Device type '{device_type}' not supported")
    api_class: type[VodafoneStationCommonApi] = class_registry[device_type]

    return api_class(
        url,
        data["username"],
        data["password"],
        session,
    )


async def get_device_type(
    host: str,
    session: ClientSession,
) -> tuple[DeviceType, URL]:
    """Find out the device type of a Vodafone Stations and returns it as enum.

    The Technicolor devices always answer with a valid HTTP response, the
    Sercomm returns 404 on a missing page. This helps to determine which we are
    talking with.
    For detecting the Sercomm devices, a look up for a CSRF token is used.

    Args:
    ----
        host (str): The router's address, e.g. `192.168.1.1`
        session (ClientSession): the client session for HTTP requests

    Returns:
    -------
    [
        device_type:
            returns `Sercomm`, `Technicolor` or raises `ModelNotSupported`
        url:
            full router url with scheme and host, e.g. `http://192.168.1.1`
    ]

    """
    for device_info in DEVICES_SETTINGS.values():
        api_path = device_info.get("login_url")
        for protocol in ["https", "http"]:
            try:
                return_url = URL(f"{protocol}://{host}")
                url = return_url.joinpath(api_path)
                _LOGGER.debug("Trying url %s", url)
                async with session.get(
                    url,
                    headers=HEADERS,
                    allow_redirects=False,
                    params=device_info.get("params"),
                    ssl=False,
                ) as response:
                    _LOGGER.debug("Response for url %s: %s", url, response.status)
                    if response.status != HTTPStatus.OK:
                        continue

                    response_text = await response.text()
                    response_json: dict[str, Any] = {}
                    if response.content_type == "application/json":
                        try:
                            response_json = orjson.loads(response_text)
                        except orjson.JSONDecodeError:
                            _LOGGER.debug("Failed to decode JSON response from %s", url)

                    if "data" in response_json and "ModelName" in response_json["data"]:
                        _LOGGER.debug(
                            "Detected device type: %s", DeviceType.TECHNICOLOR
                        )
                        return (DeviceType.TECHNICOLOR, return_url)

                    if "X_VODAFONE_ServiceStatus_1" in response_json:
                        session.cookie_jar.clear()  # Needed to cleanup session
                        _LOGGER.debug("Detected device type: %s", DeviceType.ULTRAHUB)
                        return (DeviceType.ULTRAHUB, return_url)

                    if "var csrf_token = " in response_text:
                        _LOGGER.debug("Detected device type: %s", DeviceType.SERCOMM)
                        return (DeviceType.SERCOMM, return_url)

            except (
                ClientConnectorSSLError,
                ClientConnectorError,
            ):
                _LOGGER.debug("Unable to login using protocol %s", protocol)
                continue

    raise ModelNotSupported



"""UltraHub Vodafone Station model API implementation."""

import base64
import contextlib
import os
from datetime import UTC, datetime, timedelta
from http import HTTPMethod, HTTPStatus
from typing import Any, cast

import orjson
from aiohttp import (
    ClientResponse,
    ClientResponseError,
    ClientSession,
    ClientTimeout,
)
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers.aead import AESCCM
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from yarl import URL

from aiovodafone.api import VodafoneStationCommonApi, VodafoneStationDevice
from aiovodafone.const import _LOGGER, DEFAULT_TIMEOUT, DEVICES_SETTINGS
from aiovodafone.exceptions import (
    AlreadyLogged,
    CannotAuthenticate,
    GenericLoginError,
    GenericResponseError,
)


class VodafoneStationUltraHubApi(VodafoneStationCommonApi):
    """Queries Vodafone Ultra Hub."""

    def __init__(
        self, url: URL, username: str, password: str, session: ClientSession
    ) -> None:
        """Initialize id as it may change in the future."""
        super().__init__(url, username, password, session)
        self.id = DEVICES_SETTINGS["UltraHub"]["default_id"]

    async def login(self, force_logout: bool = False) -> bool:
        """Router login."""
        _LOGGER.debug("Logging into %s", self.base_url.host)

        if not force_logout:
            self.csrf_token = ""

        reply = await self._auto_hub_request_page_result(
            HTTPMethod.GET,
            DEVICES_SETTINGS["UltraHub"]["login_url"],
            params={"X_INTERNAL_FIELDS": "X_RDK_ONT_Veip_1_OperationalState"},
        )

        reply_json = await reply.json()

        if "X_INTERNAL_ID" in reply_json:
            self.id = reply_json["X_INTERNAL_ID"]
            self.session.cookie_jar.update_cookies(reply.cookies)

        if self.csrf_token == "":
            raise CannotAuthenticate

        reply = await self._auto_hub_request_page_result(
            HTTPMethod.GET,
            "api/users/details.jst",
            params={"__id": self.id, "X_INTERNAL_FIELDS": "X_VODAFONE_WebUISecret"},
        )

        reply_json = await reply.json()

        if "X_VODAFONE_WebUISecret" in reply_json:
            web_secret = reply_json["X_VODAFONE_WebUISecret"]
            salt_web_ui = web_secret[:10]
            salt = web_secret[10:]
            password = await self._encrypt_string(salt, salt_web_ui)
            payload = {
                "__id": self.id,
                "X_VODAFONE_Password": password,
                "Push": str(force_logout).lower(),
                "csrf_token": self.csrf_token,
            }

            reply = await self._auto_hub_request_page_result(
                HTTPMethod.POST,
                "api/users/login.jst",
                payload=payload,
            )

            self.session.cookie_jar.update_cookies(reply.cookies)
            reply_json = await reply.json()

            if reply_json.get("X_INTERNAL_Password_Status") == "Invalid_PWD":
                await self._cleanup_session()
                raise CannotAuthenticate

            if reply_json.get("X_INTERNAL_Is_Duplicate") == "true":
                await self._cleanup_session()
                raise AlreadyLogged

            return True

        raise GenericLoginError

    async def _encrypt_string(
        self,
        salt: str,
        salt_web_ui: str,
    ) -> str:
        """Calculate login hash (password), the salt and the salt (web UI).

        Args:
        ----
            salt (str): salt given by the login response
            salt_web_ui (str): salt given by the web UI

        Returns:
        -------
            str: the hash for the session API

        """
        _LOGGER.debug("Calculate credential hash")

        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=16,
            salt=bytes(salt, "utf-8"),
            iterations=1000,
        )

        key = kdf.derive(bytes(salt_web_ui, "utf-8"))

        iv = os.urandom(16)
        nonce = self._truncate_iv(iv, len(self.password) * 8, 8)
        aesccm = AESCCM(key, 8)
        ct = aesccm.encrypt(nonce, bytes(self.password, "utf-8"), None)
        b64_ct = base64.b64encode(ct).decode("ascii").strip()
        b64_iv = base64.b64encode(iv).decode("ascii").strip()

        value_dict = {
            "iv": b64_iv,
            "v": 1,
            "iter": 1000,
            "ks": 128,
            "ts": 64,
            "mode": "ccm",
            "adata": "",
            "cipher": "aes",
            "ct": b64_ct,
        }
        return cast("str", orjson.dumps(value_dict).decode("utf-8"))

    def _truncate_iv(
        self,
        iv: bytes,
        ol: int,  # in bits (output length including tag)
        tlen: int,  # in bytes
    ) -> bytes:
        """Calculate the nonce as it can not be 16 bytes."""
        ivl = len(iv)  # iv length in bytes
        ol = (ol - tlen) // 8

        # "compute the length of the length" (see ccm.js)
        loop = 2
        max_length_field_bytes = 4  # Maximum L parameter per CCM spec
        while (loop < max_length_field_bytes) and (ol >> (8 * loop)) > 0:
            loop += 1
        loop = max(loop, 15 - ivl)

        return iv[: (15 - loop)]

    async def _auto_hub_request_page_result(
        self,
        method: str,
        page: str,
        payload: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
        timeout: ClientTimeout = DEFAULT_TIMEOUT,
    ) -> ClientResponse:
        """Request data from a web page."""
        url = self.base_url.joinpath(page)
        _LOGGER.debug("%s page %s", method, url)

        try:
            response = await self.session.request(
                method,
                url,
                params=params,
                data=payload,
                headers=self.headers,
                allow_redirects=False,
                timeout=timeout,
                ssl=False,
            )
            if response.status != HTTPStatus.OK:
                _LOGGER.warning(
                    "%s page %s failed: %s",
                    method,
                    url,
                    response.status,
                )
                raise GenericResponseError
        except ClientResponseError as err:
            _LOGGER.exception(
                "%s page %s from host %s failed", method, page, self.base_url.host
            )
            raise GenericResponseError from err
        else:
            reply_json = await response.json()
            if "csrf_token" in reply_json:
                self.csrf_token = reply_json["csrf_token"]

            return response

    async def _cleanup_session(self) -> None:
        """Cleanup session."""
        self.csrf_token = ""
        self.session.cookie_jar.clear()  # may be clear by domain

    def convert_uptime(self, uptime: str) -> datetime:
        """Convert uptime to datetime."""
        return datetime.now(tz=UTC) - timedelta(
            seconds=int(uptime),
        )

    async def get_devices_data(self) -> dict[str, VodafoneStationDevice]:
        """Get router device data."""
        _LOGGER.debug("Get hosts")

        devices_dict: dict[str, VodafoneStationDevice] = {}

        reply = await self._auto_hub_request_page_result(
            HTTPMethod.GET, "api/device/bulk/details.jst"
        )

        reply_json = await reply.json()

        for device in reply_json.get("hosts", []):
            connected = device["Active"] == "true"
            connection_type = (
                "WiFi" if "WiFi" in device["Layer1Interface"] else "Ethernet"
            )
            ip_address = device["IPv4Address_1_IPAddress"]
            name = device["HostName"]
            mac = device["PhysAddress"]
            dev_type = device["X_VODAFONE_Fingerprint_Class"]
            wifi = device["X_CISCO_COM_RSSI"]

            vdf_device = VodafoneStationDevice(
                connected=connected,
                connection_type=connection_type,
                ip_address=ip_address,
                name=name,
                mac=mac,
                type=dev_type,
                wifi=wifi,
            )
            devices_dict[mac] = vdf_device

        return devices_dict

    async def get_sensor_data(self) -> dict[str, Any]:
        """Get router sensor data."""
        reply = await self._auto_hub_request_page_result(
            HTTPMethod.GET, "api/device/details.jst"
        )

        reply_json = await reply.json()

        data: dict[str, Any] = {}

        data["sys_firmware_version"] = reply_json["SoftwareVersion"]
        data["sys_hardware_version"] = reply_json["HardwareVersion"]
        data["sys_serial_number"] = reply_json["SerialNumber"]
        data["sys_uptime"] = reply_json["UpTime"]
        data["wan_status"] = ""
        data["cm_status"] = ""
        data["lan_mode"] = reply_json["X_VODAFONE_WANType"]

        for device in reply_json["INTERNAL_CPEInterface_List"]:
            if device["DisplayName"] == "WWAN":
                data["wan_status"] = device["Phy_Status"]
            if device["DisplayName"] == "WANoE":
                data["cm_status"] = device["Phy_Status"]

        return data

    async def get_docis_data(self) -> dict[str, Any]:
        """Get router docis data."""
        return {}

    async def get_voice_data(self) -> dict[str, Any]:
        """Get router voice data."""
        return {}

    async def restart_connection(self, connection_type: str) -> None:  # noqa: ARG002
        """Internet Connection restart."""
        msg = f"Method not implemented for UltraHub device {self.base_url.host}"
        _LOGGER.error(msg)
        raise NotImplementedError(msg)

    async def restart_router(self) -> None:
        """Router restart."""
        _LOGGER.debug("Restarting router %s", self.base_url.host)

        payload = {"RebootDevice": "true", "csrf_token": self.csrf_token}

        with contextlib.suppress(GenericResponseError):
            await self._auto_hub_request_page_result(
                HTTPMethod.POST, "api/device/update.jst", payload=payload
            )

        await self._cleanup_session()

    async def logout(self) -> None:
        """Router logout."""
        _LOGGER.debug("Log out of router %s", self.base_url.host)
        if hasattr(self, "session") and self.csrf_token != "":
            payload = {"__id": self.id, "csrf_token": self.csrf_token}

            with contextlib.suppress(GenericResponseError):
                await self._auto_hub_request_page_result(
                    HTTPMethod.POST, "api/users/logout.jst", payload=payload
                )

            await self._cleanup_session()



"""Constants for Vodafone Station."""

import logging
from typing import Any

from aiohttp import ClientTimeout

_LOGGER = logging.getLogger(__package__)

HEADERS = {
    "User-Agent": (
        "Mozilla/5.0 (X11; Fedora; Linux x86_64; rv:78.0) Gecko/20100101 Firefox/78.0"
    ),
    "Accept-Language": "en-GB,en;q=0.5",
    "X-Requested-With": "XMLHttpRequest",
    "Priority": "u=1",
}
DEFAULT_TIMEOUT = ClientTimeout(10)
POST_RESTART_TIMEOUT = ClientTimeout(2)

DEVICES_SETTINGS: dict[str, Any] = {
    "Sercomm": {
        "login_url": "login.html",
        "total_fields_num": 8,
    },
    "Technicolor": {
        "login_url": "api/v1/login_conf",
        "user_already_logged_in": "MSG_LOGIN_150",
    },
    "UltraHub": {
        "login_url": "api/config/details.jst",
        "params": {"X_INTERNAL_FIELDS": "X_VODAFONE_ServiceStatus_1"},
        "default_id": 3,
    },
}

DEVICE_SERCOMM_LOGIN_STATUS = [
    "not logged",
    "logged",
    "already logged",
    "credential error",
    "credential error",
    "password mismatch",
    "incorrect challenge",
    "password mismatch",
]
